package jobqueue

import (
	"gorm.io/gorm"
	"sync"
)

type DBSaverStatus int

const (
	DBSaverStatusStopped = iota
	DBSaverStatusRunning
)

type DBSaver struct {
	saveChannel chan GenericJob
	stopChan    chan struct{}
	migrateFlag bool
	manager     *Manager
	status      DBSaverStatus
	mu          sync.Mutex
}

type RunnerData string
type SchedulerData string

// NewDBSaver creates a new DBSaver
func NewDBSaver() *DBSaver {
	return &DBSaver{
		saveChannel: make(chan GenericJob, 100),
		stopChan:    make(chan struct{}),
	}
}

// SetManager sets the manager of the DBSaver
func (s *DBSaver) SetManager(manager *Manager) *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.manager = manager
	return s
}

func (s *DBSaver) setStatus(status DBSaverStatus) *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.status = status
	return s
}

// isStatus returns true if the DBSaver has the given status
func (s *DBSaver) isStatus(status DBSaverStatus) bool {
	s.mu.Lock()
	defer s.mu.Unlock()

	return s.status == status

}

// Start starts the DBSaver
func (s *DBSaver) Start() error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	if s.isStatus(DBSaverStatusRunning) {
		return nil
	}

	db := s.manager.database

	if !s.migrateFlag {
		err := db.AutoMigrate(&JobPersistence{}, &JobLog{}, &JobStats{})
		if err != nil {
			return err
		}
		s.migrateFlag = true
	}

	go func() {
		s.setStatus(DBSaverStatusRunning)
		defer func() {
			s.setStatus(DBSaverStatusStopped)
		}()
		for {
			select {
			case job := <-s.saveChannel:

				err := db.Transaction(func(tx *gorm.DB) error {

					permJob := job.GetPersistence()

					memLogs := permJob.Logs
					permJob.Logs = nil

					result := tx.FirstOrCreate(&permJob, map[string]interface{}{"id": permJob.GetID()}).
						Assign(permJob)

					err := result.Error

					// Find or create JobMeta
					if err != nil {
						s.logError("Error while saving job", "error", err)
					}

					if result.RowsAffected == 0 {
						tx.Model(&permJob.Stats).Updates(permJob.Stats)
					}

					for _, log := range memLogs {
						log.LogID = 0
						_ = tx.Create(&log)
						// no error handling, if it fails, it fails
					}

					return nil // Commit the transaction
				})

				if err != nil {
					s.logError("Error while saving job", "error", err)
				}

			case <-s.stopChan:
				return
			}
		}
	}()

	return nil
}

// logError logs an error
func (s *DBSaver) logError(msg string, keysAndValues ...interface{}) {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.logger == nil {
		return
	}

	s.manager.logger.Error(msg, keysAndValues...)
}

// Stop stops the DBSaver
func (s *DBSaver) Stop() *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.stopChan <- struct{}{}
	return s
}

// SaveJob saves a job to the database
func (s *DBSaver) SaveJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.saveChannel == nil {
		return ErrDBSaverNotInitialized
	}

	if s.status != DBSaverStatusRunning {
		return ErrDBSaverNotRunning
	}

	s.saveChannel <- job
	return nil
}
